#include "Stream_Archive.h"
#include "ntp.h"

#include "ace/Time_Value.h"
#include "ace/Timer_Queue.h"
#include "ace/INET_Addr.h"
#include "ace/OS.h"

#include "Record_Archive_Manager.h"
#include <stdio.h>
#include <sys/stat.h>
#include <stream.h>

#define USE_STDIO_BUFFERING

#define NO_MMFS

#ifndef _AIX
#define NO_MMFS
#endif

#ifndef NO_MMFS
#include "TigerShark.h"
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
//  Stream_Archive:  Write RTP packets out to a file for a single data stream.  Also
//                              write a db info file that describes the packet data
//
// Modification History
//  16 jan 98:  lisa - adapted from SSRC_Writer
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Constants
static const short kMaxCNameLen = 256;  // The maximum number of CNAME characters


Stream_Archive::Stream_Archive( Record_Archive_Manager* archiveMgr , long ssrc  ,
				VOY_Database_ptr db , long sessionId ,
				const char* streamId ) :
    archiveMgr_( archiveMgr ) , ssrc_( ssrc ) , db_( db ) , sessionId_( sessionId ) , lastRtpSeq( 0 ) ,
    bytes_( 0 ) , totalBytes_( 0 ) , fp_( NULL ) , infoFp_( NULL ) , writingOutput_( 0 ) ,
    haveSavedRtcpPacket_( 0 ) , missedPackets_( 0 ) , packetsSeen_( 0 ) , fileSize_( 0 ) ,
    totalMissedPackets_( 0 ) , totalPacketsSeen_( 0 ) , haveSynchInfo_( 0 )
{
    printf("Stream_Archive::Stream_Archive:  for ssrc %x\n", ssrc_);

    streamId_ = ACE_OS::strdup( streamId );

    fbHaveCName = 0; // Indicate that we have not yet retrieved the CNAME

    const VOY_String& directory = archiveMgr->directory( );

    if( directory.length( ) )
    {
	const char* dir = directory.c_str( );
	streamFilename_ = form( "%s/%x" , dir , ssrc_ );
	infoFilename_ = form( "%s/%x_info" , dir , ssrc_ );
    }
    else
    {
	// exception
	assert( 0 );
    }

    printf( " Stream_Archive::Stream_Archive: built filenames %s and %s\n" ,
	    streamFilename_.c_str( ) , infoFilename_.c_str( )  );
    
    bzero( typesSeen_ , sizeof( typesSeen_ ) );
}

Stream_Archive::~Stream_Archive( )
{
    terminate( );
    //printf("Stream_Archive::~Stream_Archive:  destructing ssrc %x\n", ssrc_);
}

void
Stream_Archive::terminate( void )
{
    //printf("Stream_Archive::terminate:   terminating ssrc %x\n" , ssrc_ );
    if( fp_ )
	fclose( fp_ );
    fp_ = NULL;

    writeFinalInfo( );
}

/*
 * Spit out info we know about the start of the stream.
 *
 * SSRC
 * starting time
 */
void
Stream_Archive::writeInitialInfo( void )
{
    ACE_Time_Value duration;

    infoFp_ = fopen( infoFilename_.c_str( ) , "w" );
    //printf( "Stream_Archive::writeInitialInfo: info filename %s\n" ,  infoFilename_.c_str( ) );
    if( infoFp_ == ( FILE* )NULL )
    {
	perror( "Stream_Archive::writeInitialInfo:  info filename error on open" );
	exit( 1 );
    }
    //char tmpBuf[ 256 ];
    
    sprintf( tmpBuf , "%x" , ssrc_ );
    setStringStreamKeyValue( "ssrc" , tmpBuf );

    sprintf( tmpBuf , "%d.%06d" , startTime_.sec( ) , startTime_.usec( ) );
    setStringStreamKeyValue( "start_time" , tmpBuf );
}

/*
 * Spit out the end-of-stream summary information:
 *
 * Duration
 */
void
Stream_Archive::writeFinalInfo( void )
{
    ACE_Time_Value duration;

    //printf( "Stream_Archive::writeFinalInfo: info filename %s\n" ,  infoFilename_.c_str( ) );
    if( infoFp_ == ( FILE* )NULL )
    {
	printf( "Stream_Archive::writeFinalInfo: returning w/o without writing info file for ssrc %ld\n" , ssrc_ );
	return;
    }
    
    duration = ACE_OS::gettimeofday( ) - startTime_;

    //char tmpBuf[ 256 ];
    short qFirstPacketType = 1;
    for( int i=0  ;  i < 128  ;  i++ )
    {
	if( typesSeen_[ i ] )
	{
	    if( !qFirstPacketType )
		sprintf( tmpBuf , " %d", i );
	    else
	    {
		qFirstPacketType = 0;
		sprintf( tmpBuf , "%d", i );
	    }
	}
    }
    setStringStreamKeyValue( "packet_types" , tmpBuf );

    sprintf( tmpBuf , "%d.%06d" , duration.sec( ) , duration.usec( ) );
    setStringStreamKeyValue( "duration" , tmpBuf );
    
    sprintf( tmpBuf , "%d" , totalPacketsSeen_ );
    setStringStreamKeyValue( "packets_seen" , tmpBuf );
    
    sprintf( tmpBuf , "%d" , totalMissedPackets_ );
    setStringStreamKeyValue( "missed_packets" , tmpBuf );
    
    sprintf( tmpBuf , "%d" , totalBytes_ );
    setStringStreamKeyValue( "bytes" , tmpBuf );

    sprintf( tmpBuf , "%ld" , fileSize_ );
    setStringStreamKeyValue( "file_size" , tmpBuf );
    
    fclose( infoFp_ );
}


void
Stream_Archive::openFile( void )
{
    struct stat s;

    fp_ = fopen( streamFilename_.c_str( ) , "w" );

    if( fp_ == ( FILE* )NULL )
    {
	perror( "Stream_Archive::open_file open: No file pointer!" );
	// exception
	exit(1);
    }

    if( fstat( fileno( fp_ ) , &s ) < 0 )
    {
	perror( "fstat" );
	return;
    }

#ifdef USE_STDIO_BUFFERING
    printf( "Stream_Archive::openFile: setting stdio blocksize to %d\n" , s.st_blksize );
    fpBuffer_ = new char[ s.st_blksize ];

    if( setvbuf( fp_ , fpBuffer_ ,  _IOFBF , s.st_blksize ) != 0 )
    {
	perror( "setvbuf" );
    }
#else
    fpBuffer_ = NULL;
    printf( "Stream_Archive::openFile: disabling stdio buffering\n" );
    if( setvbuf( fp_ , NULL , _IONBF, 0 ) != 0 )
    {
	perror( "setvbuf" );
    }
#endif
    
#ifndef NO_MMFS
    if( s.st_vfstype == VFSTYPE_MMFS )
    {
	int ival;

	ival = 5000000;

#ifdef OLD_MMFS
	
	if( mmfs_qioctl( fileno( fp_ ) , SET_RATE , &ival ) < 0 )
	    perror( "SET_RATE" );

	if( mmfs_qioctl( fileno(fp_) , GET_RATE , &ival ) < 0 )
	    perror( "GET_RATE" );
	else
	    printf( "Stream_Archive::openFile: rate=%d\n", ival );

	ival = TS_BUFFERS;
	if( mmfs_qioctl( fileno( fp_ ) , SET_BUFFERING , &ival ) < 0 )
	    perror( "SET_BUFFERING" );
#else
	if( tsfattr( fileno( fp_ ) , SET_RATE , &ival , NULL ) < 0 )
	    perror( "SET_RATE" );

	if( tsfattr( fileno( fp_ ) , GET_RATE , &ival , NULL ) < 0 )
	    perror( "GET_RATE" );
	else
	    printf( "Stream_Archive::openFile: rate=%d\n" , ival );

	ival = TS_BUFFERS;
	if( tsfattr( fileno( fp_ ) , SET_BUFFERING , &ival , NULL ) < 0 )
	    perror( "SET_BUFFERING" );
#endif
	
    }
    else
    {
	printf( "Stream_Archive::openFile: warning: %s not in MMFS\n" , streamFilename_  );
	return;
    }
#endif
}

void
Stream_Archive::handlePacket( RTP_Packet::RTP_Header* packetHeader , char* packet ,
			      int packetSize ,	ACE_INET_Addr& packetSource ,
			      int initialPacket )
{
    /*
     * If we aren't writing output yet, we need to start.
     * Open the file, write the header, and write the saved
     * RTCP packet if there is one.
     */
	 
    if( !writingOutput_ && archiveMgr_->recording( ) )
    {
	//char tmpBuf[ 256 ];
	writeFileHeader( packetSource );
	writeInitialInfo( );
	if( haveSynchInfo_ )
	{
	    sprintf( tmpBuf , "%f" , syncNtpTimestamp_ );
	    setStringStreamKeyValue( "synch_ntp_timestamp" , tmpBuf );
	    sprintf( tmpBuf , "%u" , syncRtpTimestamp_ );
	    setStringStreamKeyValue( "synch_rtp_timestamp" , tmpBuf );
	}
	if( haveSavedRtcpPacket_ )
	{
	    int len;
	    char *buf;
	    
	    buf = savedRtcpPacket_.get_buffer( &len );
	    writePacket( buf , savedRtcpPacketSize_ , 1 );
	}

	/*
	 * Write out the timestamp for this first RTP packet in the
	 * file.
	 */
	sprintf( tmpBuf , "%u" ,  packetHeader->get_timestamp( ) );
	setStringStreamKeyValue( "starting_timestamp" , tmpBuf );
	
	writingOutput_ = 1;
    }

    bytes_ += packetSize;
    totalBytes_ += packetSize;

    packetsSeen_++;
    totalPacketsSeen_++;
    lastRtpSeq++;
    if( packetHeader->get_sequence( ) != lastRtpSeq && lastRtpSeq != 65536 )
    {
	printf( "Stream_Archive::handlePacket(%x): missed packet %d\n" , ssrc_ , lastRtpSeq );
	missedPackets_++;
	totalMissedPackets_++;
	lastRtpSeq = packetHeader->get_sequence( );
    }

    writePacket( packet , packetSize , 0 );

    int packetType = packetHeader->get_packet_type( );
    if( !typesSeen_[ packetType] )
    {
	//char tmpBuf[ 256 ];
	typesSeen_[ packetType ] = 1;
	sprintf( tmpBuf , "%d" , packetType );
	setStringStreamKeyValue( "packetTypesSeen" , tmpBuf );
    }

    if( !haveSynchInfo_ )
    {
	//printf("Stream_Archive::handlePacket: flags=%x, v=%d, pad=%d, ext=%d, cc=%d, marker=%d, pt=%d, seq=%d, ssrc=%x, ts=%u\n",
	//     packetHeader->get_flags( ) , packetHeader->get_version( ) ,
	//     packetHeader->get_padding( ) , packetHeader->get_extension( ) ,
	//     packetHeader->get_csrc_count( ) , packetHeader->get_marker( ) ,
	//     packetHeader->get_packet_type( ) , packetHeader->get_sequence( ) ,
	//     packetHeader->get_ssrc( ) , packetHeader->get_timestamp( ) );
    }
}
    
void
Stream_Archive::handlePacket( RTCP_Packet::RTCP_Header* packetHeader , char* packet ,
			      int packetSize , ACE_INET_Addr& packetSource ,
			      int initialPacket )
{
    if( !writingOutput_ )
    {
	char *buf;
	int len;
	buf = savedRtcpPacket_.get_buffer( &len );
	bcopy( packet , buf , packetSize );
	savedRtcpPacketSize_ = packetSize;
	haveSavedRtcpPacket_ = 1;
    }
    else
    {
	writePacket( packet , packetSize , 1 );
    }

}

void Stream_Archive::handleSubpacket(RTCP_Packet::RTCP_Header* packetHeader ,
				      char* packet , int packetSize ,
				      ACE_INET_Addr& packetSource )
{
    if( writingOutput_ )
    {
	switch( packetHeader->get_packet_type( ) )
	{
	case RTCP_Packet::PT_BYE:
	case RTCP_Packet::PT_BYE_OLD:	
	    archiveMgr_->endOfStream( this );
	    break;

	case RTCP_Packet::PT_SDES_OLD:
	case RTCP_Packet::PT_SDES:

	    // Extract the CNAME from the source description block (SDES) of an RTCP packet.
	    // CNAME is the canonical name of the participant in the form <userid>@nnn.nnn.nnn.nnn.
	    // The SDES items that follow the header have the following format:
	    //  1st 8 bits: SDES item id, 2nd 8 bits: item length in bytes, following bits: n bytes of data
	    if( !fbHaveCName )
	    {
		// Position the pointer to the beginning of the SDES item
		char* pSDES = packet + sizeof( RTCP_Packet::RTCP_Header );
 
		if( pSDES[ 0 ] == 1 ) // There are several SDES blocks. CName==1
		{
		    short length = pSDES[ 1 ] < kMaxCNameLen ? pSDES[ 1 ] :
			kMaxCNameLen-1;
		
		    // Extract the name
		    char cName[ kMaxCNameLen ];
		    strncpy( cName , &pSDES[ 2 ] , length );
		    cName[ length ] = '\0';
		
		    //char tmpBuf[ 256 ];
		    sprintf( tmpBuf ,  "%s" , cName );
		    setStringStreamKeyValue( "sdes_cname" , tmpBuf );
		    fbHaveCName = 1; // Name retrieved.  No need to hunt for it again.
		}
	    }

	    break;
	
	case RTCP_Packet::PT_SR:
	case RTCP_Packet::PT_SR_OLD:

	    /*
	     * If we aren't writing output yet, remember the ssrc &
	     * timestamp for this SR. We will use this on playback
	     * to precisely recreate the timing of the original stream
	     * (by remembering the mapping between wallclock time and
	     * rtp timestamp).
	     */
	    if( !haveSynchInfo_ )
	    {
		RTCP_Packet::Sender_Info* si;

		si = ( RTCP_Packet::Sender_Info * )( packet + 8 );

		double d = NTP_TO_GETTIMEOFDAY( ntohl(si->ntp_timestamp_msw_ ) ,
						ntohl( si->ntp_timestamp_lsw_ ) );
	    
		//printf("Stream_Archive::handleSubpacket: ssrc=%x, ntp %u, %u, d=%f, rtp=%x, pack=%d, oct=%d\n" ,
		//     packetHeader->get_ssrc( ) , ntohl( si->ntp_timestamp_msw_ ) ,
		//     ntohl( si->ntp_timestamp_lsw_ ) , d , ntohl( si->rtp_timestamp_ ) ,
		//     ntohl( si->packet_count_ ) , ntohl( si->octet_count_ ) );

		//char tmpBuf[ 256 ];
		sprintf( tmpBuf , "%f" , d );
		setStringStreamKeyValue( "synch_ntp_timestamp" , tmpBuf );
		sprintf( tmpBuf , "%u" , ntohl(si->rtp_timestamp_ ) );
		setStringStreamKeyValue( "synch_rtp_timestamp" , tmpBuf );
		
		syncNtpTimestamp_ = d;
		syncRtpTimestamp_ = ntohl( si->rtp_timestamp_ );
	    
		haveSynchInfo_ = 1;

		break;
	    }
	}
    }
}

void Stream_Archive::writeFileHeader(ACE_INET_Addr& source)
{
    RTP_Dump_Header rdHeader_;

    startTime_ = ACE_OS::gettimeofday();

    //printf( "Stream_Archive::writeFileHeader(%x)\n" , ssrc_  );

    openFile( );

    rdHeader_.start_s = htonl( startTime_.sec( ) );
    rdHeader_.start_us = htonl( startTime_.usec( ) );
    rdHeader_.source = htonl( source.get_ip_address( ) );
    rdHeader_.port = htons( source.get_port_number( ) );

    if( fwrite( ( unsigned char* )&rdHeader_ , sizeof( rdHeader_ ) , 1 , fp_ ) != 1)
    {
	perror( "Stream_Archive::writeFileHeader: write failed" );
	exit( 1 );
    }
    else
	fileSize_ += sizeof(rdHeader_ );

    startMs_ = startTime_.sec() * 1000 + startTime_.usec() / 1000;
}

void Stream_Archive::writePacket( char* packet , int length , int type )
{
    if( !archiveMgr_->recording( ) )
	return;
    
    ACE_Time_Value now = ACE_OS::gettimeofday( );

    rdPacket_.length = htons (length + sizeof( rdPacket_ ) );
    rdPacket_.type = htons( type );
    rdPacket_.offset = htonl( now.sec( ) * 1000 + now.usec( )  / 1000 - startMs_ );
    if( fwrite( &rdPacket_ , sizeof( rdPacket_ ) , 1 , fp_ ) != 1 )
    {
	perror( "Stream_Archive::writePacket: write header failed" );
	exit( 1 );
    }
    else
	fileSize_ += sizeof(rdPacket_ );
	

    if( fwrite( ( u_char* )packet , length , 1 , fp_ ) != 1 )
    {
	perror( "Stream_Archive::writePacket: write data failed" );
	exit( 1 );
    }
    else
	fileSize_ += length;
}
 
void Stream_Archive::tick(const ACE_Time_Value& this_tick ,
			  const ACE_Time_Value& last_tick ,
			  const ACE_Time_Value& delta )
{
#ifdef STATS
    double bw = 0.0;

    bw = 8.0 * ( double )bytes_  /  ( ( double )delta.sec( ) + ( double )delta.usec( ) / 1000000.0 );
    //printf( "Stream_Archive::tick%d bytes in %d.%06d sec = %f Mb/s %d/%d missed packets (%5.2f%%)\n", bytes_ , delta.sec( ) , delta.usec( ) , bw / 1048576.0 , missedPackets_ , packetsSeen_ , 100 * ( double )missedPackets_  /  ( double )packetsSeen_ );
#endif
    bytes_ = 0;
    missedPackets_ = packetsSeen_ = 0;
}

void
Stream_Archive::setStringStreamKeyValue( char* key , char* value )
{
    if( infoFp_ )
    {
	//printf( "Stream_Archive::setStringStreamKeyValue: key %s, value ...%s...\n" , key , value );
	db_->setStreamKeyValue( sessionId_ , streamId_ , key , value );
	fprintf( infoFp_ , "%s: %s\n" , key , value );
	long err = fflush( infoFp_ );
    }
}
